from collections import defaultdict
from itertools import chain

import numpy as np

import networkx as nx
from networkx.algorithms.shortest_paths.weighted import single_source_dijkstra

from Bio.PDB.kdtrees import KDTree

from luna.pdb.io.selector import Selector, AtomSelector
from luna.pdb.convert import biopython_entity_to_mol

from luna.interaction.contact import get_proximal_compounds
from luna.interaction.contact import get_contacts_with
from luna.interaction.type import InteractionType
from luna.mol.atom import ExtendedAtom, AtomData
from luna.mol.precomp_data import DefaultResidueData
from luna.mol.charge_model import OpenEyeModel
from luna.mol.features import ChemicalFeature
from luna.wrappers.base import MolWrapper
from luna.util.exceptions import MoleculeSizeError, IllegalArgumentError
from luna.util.default_values import COV_SEARCH_RADIUS, METAL_COMPLEX_DIST
from luna.util import math as im
from luna.util.file import pickle_data, unpickle_data
from luna.version import __version__

import logging
logger = logging.getLogger()


SS_BOND_FEATURES = ["Atom", "Acceptor", "ChalcogenDonor", "Hydrophobic"]

DEFAULT_RES_DATA = DefaultResidueData()


class AtomGroupsManager():
    """Store and manage `AtomGroup` objects.

    Parameters
    ----------
    atm_grps : iterable of `AtomGroup`, optional
        An initial sequence of `AtomGroup` objects.
    entry : :class:`~luna.mol.entry.Entry`, optional
        The chain or molecule from where the atom groups were perceived.

    Attributes
    ----------
    entry : :class:`~luna.mol.entry.Entry`
        The chain or molecule from where the atom groups were perceived.
    graph : :py:class:`networkx.Graph`
        Represent ``entry`` as a graph and its vicinity.
    version : str
        The LUNA version when the object was created.
    """

    def __init__(self, atm_grps=None, entry=None):
        self._atm_grps = []
        self.entry = entry
        self._child_dict = {}

        self.graph = nx.Graph()

        self.version = __version__

        self._compounds = set()

        self.add_atm_grps(atm_grps)

    @property
    def atm_grps(self):
        """iterable of `AtomGroup`, read-only: The sequence of `AtomGroup`
        objects. Additional objects should be added using the method
        :py:meth:`add_atm_grps`."""
        return self._atm_grps

    @property
    def compounds(self):
        """set of :class:`~luna.pdb.core.residue.Residue`, read-only:
        Compounds comprising the ``atm_grps``."""
        return self._compounds

    @property
    def child_dict(self):
        """dict, read-only: Mapping between atoms (`ExtendedAtom`) and atom
        groups (`AtomGroup`).

        The mapping is a dict of {tuple of `ExtendedAtom` instances :
        `AtomGroup`} and is automatically updated when :py:meth:`add_atm_grps`
        is called."""
        return self._child_dict

    @property
    def size(self):
        """int, read-only: The number of atom groups in ``atm_grps``."""
        return len(self._atm_grps)

    @property
    def summary(self):
        """dict, read-only: The number of physicochemical features in
        ``atm_grps``."""
        summary = defaultdict(int)
        for grp in self.atm_grps:
            for feature in grp.features:
                summary[feature] += 1

        return summary

    def find_atm_grp(self, atoms):
        """ Find the atom group that contains the sequence of atoms ``atoms``.

        Returns
        -------
         : `AtomGroup` or None
            An atom group object or None if ``atoms`` is not in the
            ``child_dict`` mapping.
        """
        return self.child_dict.get(tuple(sorted(atoms)), None)

    def get_all_interactions(self):
        """Return all interactions established by the atom groups in
        ``atm_grps``.

        Returns
        -------
         : set of :class:`~luna.interactions.type.InteractionType`
            All interactions.
        """
        return set(chain.from_iterable([atm_grp.interactions
                                        for atm_grp in self.atm_grps]))

    def apply_filter(self, func):
        """Apply a filtering function over the atom groups in ``atm_grps`.

        Parameters
        ----------
        func : callable
            A filtering function that returns True case an `AtomGroup`
            object is valid and False otherwise.

        Yields
        ------
        `AtomGroup`
            A valid `AtomGroup` object.
        """
        for atm_grp in self.atm_grps:
            if func(atm_grp):
                yield atm_grp

    def filter_by_types(self, types, must_contain_all=True):
        """Filter `AtomGroup` objects by their physicochemical features.

        Parameters
        ----------
        types : iterable of str
            A sequence of physicochemical features.
        must_contain_all : bool
            If True, an `AtomGroup` object should contain all physicochemical
            features in ``types`` to be accepted. Otherwise, it will be
            filtered out.

        Yields
        ------
        `AtomGroup`
            A valid `AtomGroup` object.
        """
        for atm_grp in self.atm_grps:
            if must_contain_all:
                if set(types).issubset(set(atm_grp.feature_names)):
                    yield atm_grp
            else:
                if len(set(types) & set(atm_grp.feature_names)) > 0:
                    yield atm_grp

    def add_atm_grps(self, atm_grps):
        """Add one or more `AtomGroup` objects to ``atm_grps`` and
        automatically update ``child_dict``."""

        atm_grps = atm_grps or []

        self._atm_grps = list(set(self._atm_grps + list(atm_grps)))

        for atm_grp in atm_grps:
            self.child_dict[tuple(sorted(atm_grp.atoms))] = atm_grp
            self._compounds.update(atm_grp.compounds)
            atm_grp.manager = self

    def remove_atm_grps(self, atm_grps):
        """Remove one or more `AtomGroup` objects from ``atm_grps`` and
        automatically update ``child_dict``.

        Any recursive references to the removed objects will also be cleared.
        """
        self._atm_grps = list(set(self._atm_grps) - set(atm_grps))

        for atm_grp in atm_grps:
            # ExtendedAtom objects keep a list of all AtomGroup objects to
            # which they belong to. So, if we don't clear the references
            # directly, the AtomGroup objects will still exist in the
            # ExtendedAtom list even when they were already removed from an
            # instance of AtomGroupsManager().
            atm_grp.clear_refs()

            # Remove the atom group from the dict.
            key = tuple(sorted(atm_grp.atoms))
            if key in self.child_dict:
                del self.child_dict[key]

    def new_atm_grp(self, atoms, features=None, interactions=None):
        """Create a new `AtomGroup` object for ``atoms`` if one does not exist
        yet. Otherwise, return the existing `AtomGroup` object, and add
        any new features and interactions to it if provided.

        Parameters
        ----------
        atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
            A sequence of atoms.
        features : iterable of :class:`~luna.mol.features.ChemicalFeature`, \
                        optional
            If provided, add ``features`` to a new or an already existing
            `AtomGroup` object.
        interactions : iterable of \
                :class:`~luna.interaction.type.InteractionType`, optional
            If provided, add ``interactions`` to a new or an already existing
            `AtomGroup` object.

        Returns
        -------
         : AtomGroup
            A new or an already existing `AtomGroup` object.
        """
        key = tuple(sorted(atoms))
        features = features or []
        interactions = interactions or []

        if key in self.child_dict:
            atm_grp = self.child_dict[key]
        else:
            atm_grp = AtomGroup(atoms)
            self.add_atm_grps([atm_grp])

        if features:
            atm_grp.add_features(features)

        if interactions:
            atm_grp.add_interactions(interactions)

        return atm_grp

    def merge_hydrophobic_atoms(self, interactions_mngr):
        """Create hydrophobic islands by merging covalently bonded hydrophobic
        atoms in ``atm_grps``. Hydrophobic islands are atom groups having the
        feature `Hydrophobe`.

        Atom-atom hydrophobic interactions in ``interactions_mngr`` are also
        converted to island-island interactions.

        Parameters
        ----------
        interactions_mngr : :class:`~luna.interaction.calc.InteractionsManager`
            An :class:`~luna.interaction.calc.InteractionsManager` object from
            where hydrophobic interactions are selected and convert from
            atom-atom to island-island interactions.
        """

        # Only hydrophobic atom groups.
        hydrop_atm_grps = list(self.filter_by_types(["Hydrophobic"]))

        # Hydrophobic islands dictionary. Keys are integer values and items are
        # defined by a set of atom groups.
        hydrop_islands = defaultdict(set)

        # It stores a mapping of an atom (represented by its full id) and a
        # hydrophobic island (defined by its keys).
        atm_mapping = {}

        island_id = 0
        for atm_grp in hydrop_atm_grps:
            # Hydrophobic atoms are defined always as only one atom.
            atm = atm_grp.atoms[0]

            # Recover the groups of all neighbors of this atom (it will merge
            # all existing islands).
            nb_grps = set([atm_mapping[nbi.full_id]
                           for nbi in atm.neighbors_info
                           if nbi.full_id in atm_mapping])

            # Already there are hydrophobic islands formed by the neighbors of
            # this atom.
            if nb_grps:
                # Merge all groups of the neighbors of this atom.
                new_island = \
                    set(chain.from_iterable([hydrop_islands.pop(nb_grp_id)
                                             for nb_grp_id in nb_grps]))
                # Include this atom to the merged group.
                new_island.add(atm)

                for k in atm_mapping:
                    if atm_mapping[k] in nb_grps:
                        atm_mapping[k] = island_id

                hydrop_islands[island_id] = new_island
                atm_mapping[atm.get_full_id()] = island_id
            else:
                atm_mapping[atm.get_full_id()] = island_id
                hydrop_islands[island_id].add(atm)

            island_id += 1

        # Create AtomGroup objects for the hydrophobic islands
        for island_id in hydrop_islands:
            # It will update an existing atom group or create a new one
            # with the informed parameters.
            hydrophobe = self.new_atm_grp(hydrop_islands[island_id],
                                          [ChemicalFeature("Hydrophobe")])
            # Update the island information
            hydrop_islands[island_id] = hydrophobe

        hydrop_interactions = \
            list(interactions_mngr.filter_by_types(["Hydrophobic"]))
        island_island_inter = defaultdict(set)
        for inter in hydrop_interactions:
            src_atm = inter.src_grp.atoms[0]
            trgt_atm = inter.trgt_grp.atoms[0]

            # The two island ids are used as key.
            key = tuple(sorted([atm_mapping[src_atm.get_full_id()],
                                atm_mapping[trgt_atm.get_full_id()]]))

            island_island_inter[key].add(inter)

        interactions = set()
        for k in island_island_inter:
            island_atms = defaultdict(set)
            for inter in island_island_inter[k]:
                src_atm = inter.src_grp.atoms[0]
                trgt_atm = inter.trgt_grp.atoms[0]

                island_atms[atm_mapping[src_atm.get_full_id()]].add(src_atm)
                island_atms[atm_mapping[trgt_atm.get_full_id()]].add(trgt_atm)

            centroid1 = im.centroid(im.atom_coordinates(island_atms[k[0]]))
            centroid2 = im.centroid(im.atom_coordinates(island_atms[k[1]]))
            cc_dist = im.euclidean_distance(centroid1, centroid2)

            params = {"dist_hydrop_inter": cc_dist}

            inter = InteractionType(hydrop_islands[k[0]],
                                    hydrop_islands[k[1]],
                                    "Hydrophobic",
                                    src_interacting_atms=island_atms[k[0]],
                                    trgt_interacting_atms=island_atms[k[1]],
                                    params=params)
            interactions.add(inter)

        # Update the list of interactions with the new island-island
        # interactions.
        interactions_mngr.add_interactions(interactions)
        # Remove atom-atom hydrophobic interactions.
        interactions_mngr.remove_interactions(hydrop_interactions)

        for atm_grp in hydrop_atm_grps:
            features = [f for f in atm_grp.features if f.name != "Hydrophobic"]

            # It may happen that a atom group ends up having no feature after
            # the remotion of the feature "Hydrophobic". This is unlikely to
            # occur as all atoms (by default) will have at least the feature
            # 'Atom'. But, depending one the pharmacophore rules definition,
            # it can occur.
            atm_grp.features = features

    def get_shortest_path_length(self, src_grp, trgt_grp, cutoff=None):
        """Compute the shortest path length between two atom groups ``src_grp``
        and ``trgt_grp``.

        The shortest path between two atom groups is defined as the shortest
        path between any of their atoms, which are calculated using Dijkstra’s
        algorithm and the graph ``graph``.

        If there is not any path between ``src_grp`` and ``trgt_grp``,
        infinite is returned.

        Parameters
        ----------
        src_grp, trgt_grp : `AtomGroup`
            Two atom groups to calculate the shortest path.
        cutoff : int
            Only paths of length <= ``cutoff`` are returned.
            If None, all path lengths are considered.

        Returns
        -------
         : int or float('inf'):
            The shortest path.
        """
        shortest_path_size = float('inf')
        for src_atm in src_grp.atoms:
            for trgt_atm in trgt_grp.atoms:
                try:
                    dist, path = single_source_dijkstra(self.graph,
                                                        src_atm,
                                                        trgt_atm,
                                                        cutoff=cutoff)
                    if dist < shortest_path_size:
                        shortest_path_size = dist
                except Exception:
                    pass
        return shortest_path_size

    def save(self, output_file, compressed=True):
        """Write the pickled representation of the `AtomGroupsManager` object
        to the file ``output_file``.

        Parameters
        ----------
        output_file : str
            The output file.
        compressed : bool, optional
            If True (the default), compress the pickled representation as a
            gzip file (.gz).

        Raises
        -------
        FileNotCreated
            If the file could not be created.
        """
        pickle_data(self, output_file, compressed)

    @staticmethod
    def load(input_file):
        """Load the pickled representation of an `AtomGroupsManager` object
        saved at the file ``input_file``.

        Returns
        ----------
         : `AtomGroupsManager`
            The reconstituted `AtomGroupsManager` object, including its set of
            atom groups and interactions.

        Raises
        -------
        PKLNotReadError
            If the file could not be loaded.
        """
        return unpickle_data(input_file)

    def __len__(self):
        # Number of atom groups.
        return self.size

    def __iter__(self):
        """Iterate over children."""
        for atm_grp in self.atm_grps:
            yield atm_grp


class AtomGroup():
    """ Represent single atoms, chemical functional groups, or simply an
    arrangement of atoms as in hydrophobes.

    Parameters
    ----------
    atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
        A sequence of atoms.
    features : iterable of :class:`~luna.mol.features.ChemicalFeature`, \
                    optional
        A sequence of chemical features.
    interactions : iterable of \
                    :class:`~luna.interaction.type.InteractionType`, optional
        A sequence of interactions established by an atom group.
    recursive : bool
        If True, add the new atom group to the list of atom groups of each atom
        in ``atoms``.
    manager : `AtomGroupsManager`, optional
        The `AtomGroupsManager` object that contains this `AtomGroup` object.
    """

    def __init__(self,
                 atoms,
                 features=None,
                 interactions=None,
                 recursive=True,
                 manager=None):
        self._atoms = sorted(atoms)

        # Atom properties
        self._coords = im.atom_coordinates(atoms)
        self._centroid = im.centroid(self.coords)
        self._normal = None

        features = features or []
        self._features = sorted(features)

        self._interactions = interactions or []
        self._hash_cache = None

        self._manager = manager

        self._recursive = recursive

        if recursive:
            for atm in self.atoms:
                atm.add_atm_grps([self])

    @property
    def atoms(self):
        """iterable of :class:`~luna.mol.atom.ExtendedAtom`, read-only: \
            The sequence of atoms that belong to an atom group."""
        return self._atoms

    @property
    def compounds(self):
        """set of :class:`~luna.pdb.core.residue.Residue`, read-only: \
            The set of unique compounds that contain the atoms in ``atoms``.

        As an atom group can be formed by the union of two or more compounds
        (e.g., amide of peptide bonds), it may return more than one compound.
        """
        return set([a.parent for a in self._atoms])

    @property
    def coords(self):
        """ array-like of floats : Atomic coordinates (x, y, z) of each \
        atom in ``atoms``."""
        return self._coords

    @property
    def centroid(self):
        """ array-like of floats, read-only: The centroid (x, y, z) of the \
        atom group.

        If ``atoms`` contains only one atom, then ``centroid`` returns the same
        as ``coords``.
        """
        return self._centroid

    @property
    def normal(self):
        """array-like of floats, read-only: The normal vector (x, y, z) of \
        the points given by ``coords``."""
        if self._normal is None:
            self._normal = im.calc_normal(self.coords)
        return self._normal

    @property
    def features(self):
        """iterable of :class:`~luna.mol.features.ChemicalFeature`: \
                A sequence of chemical features.

        To add or remove a feature use :py:meth:`add_features`
        or :py:meth:`remove_features`, respectively."""
        return self._features

    @features.setter
    def features(self, features):
        self._features = sorted(features)
        # Reset hash.
        self._hash_cache = None

    @property
    def feature_names(self):
        """iterable of str: The name of each chemical feature in \
        ``features``."""
        return [f.name for f in self.features]

    @property
    def interactions(self):
        """iterable of :class:`~luna.interaction.type.InteractionType`: \
            The sequence of interactions established by an atom group.

        To add or remove an interaction use :py:meth:`add_interactions`
        or :py:meth:`remove_interactions`, respectively."""
        return self._interactions

    @interactions.setter
    def interactions(self, interactions):
        self._interactions = interactions

    @property
    def manager(self):
        """`AtomGroupsManager`: The `AtomGroupsManager` object that contains \
        an `AtomGroup` object."""
        return self._manager

    @manager.setter
    def manager(self, manager):
        if isinstance(manager, AtomGroupsManager):
            self._manager = manager
        else:
            raise IllegalArgumentError("The informed atom group manager must "
                                       "be an instance of '%s'."
                                       % AtomGroupsManager)

    @property
    def size(self):
        """int: The number of atoms comprising an atom group."""
        return len(self.atoms)

    def has_atom(self, atom):
        """Check if an atom group contains a given atom ``atom``.

        Parameters
        ----------
        atom : :class:`~luna.mol.atom.ExtendedAtom`

        Returns
        -------
         : bool
            If the atom group contains or not ``atom``.
        """
        return atom in self.atoms

    def contain_group(self, atm_grp):
        """Check if the atom group ``atm_grp`` is a subset of this atom group.

        For example, consider the benzene molecule.
        Its aromatic ring itself forms an `AtomGroup` object composed of all of
        its six atoms. Consider now any subset of carbons in the benzene
        molecule. This subset forms an `AtomGroup` object that is part of the
        group formed by the aromatic ring. Therefore, in this example,
        :meth:`contain_group` will return True because the aromatic ring
        contains the subset of hydrophobic atoms.

        Parameters
        ----------
        atm_grp : :class:`~luna.mol.groups.AtomGroup`

        Returns
        -------
         : bool
            If one atom group contains another atom group.
        """
        return set(atm_grp.atoms).issubset(set(self.atoms))

    def get_serial_numbers(self):
        """Get the serial number of each atom in an atom group."""
        return [a.get_serial_number() for a in self.atoms]

    def get_chains(self):
        """Get all unique chains in an atom group."""
        return sorted(set([a.get_parent_by_level("C").id for a in self.atoms]))

    def get_interactions_with(self, atm_grp):
        """Get all interactions that an atom group establishes with another
        atom group ``atm_grp``.

        Returns
        -------
         : iterable of :class:`~luna.interactions.type.InteractionType`
           All interactions established with the atom group ``atm_grp``.
        """
        target_interactions = []

        for inter in self.interactions:
            if inter.src_grp == atm_grp or inter.trgt_grp == atm_grp:
                target_interactions.append(inter)
        return target_interactions

    def get_shortest_path_length(self, trgt_grp, cutoff=None):
        """Compute the shortest path length between this atom group to another
        atom group ``trgt_grp``.

        The shortest path between two atom groups is defined as the shortest
        path between any of their atoms, which are calculated using
        Dijkstra’s algorithm.

        If ``manager`` is not provided, None is returned.

        If there is not any path between ``src_grp`` and ``trgt_grp``,
        infinite is returned.

        Parameters
        ----------
        trgt_grp : `AtomGroup`
            The target atom group to calculate the shortest path.
        cutoff : int, optional
            Only paths of length <= ``cutoff`` are returned.
            If None, all path lengths are considered.

        Returns
        -------
         : int, float('inf'), or None:
            The shortest path.
        """
        if self.manager is not None:
            return self.manager.get_shortest_path_length(self,
                                                         trgt_grp,
                                                         cutoff)
        return None

    def add_features(self, features):
        """ Add :class:`~luna.mol.features.ChemicalFeature` objects
        to ``features``."""
        self._features = sorted(set(self.features + list(features)))
        # Reset hash.
        self._hash_cache = None

    def remove_features(self, features):
        """ Remove :class:`~luna.mol.features.ChemicalFeature` objects
        from ``features``."""
        self._features = sorted(set(self.features) - set(features))
        # Reset hash.
        self._hash_cache = None

    def add_interactions(self, interactions):
        """ Add :class:`~luna.interaction.type.InteractionType` objects
        to ``interactions``."""
        self._interactions = list(set(self.interactions + list(interactions)))

    def remove_interactions(self, interactions):
        """ Remove :class:`~luna.interaction.type.InteractionType` objects
        from ``interactions``."""
        self._interactions = list(set(self.interactions) - set(interactions))

    def is_water(self):
        """Return True if all atoms in the atom group belong to water
        molecules."""
        return all([a.parent.is_water() for a in self.atoms])

    def is_hetatm(self):
        """Return True if all atoms in the atom group belong to hetero group,
        i.e., non-standard residues of proteins, DNAs, or RNAs, as well as
        atoms in other kinds of groups, such as carbohydrates, substrates,
        ligands, solvent, and metal ions.

        Hetero groups are designated by the flag HETATM in the PDB format."""
        return all([a.parent.is_hetatm() for a in self.atoms])

    def is_metal(self):
        """Return True if all atoms in the atom group are metal ions."""
        return all([a.parent.is_metal() for a in self.atoms])

    def is_residue(self):
        """Return True if all atoms in the atom group belong to standard
        residues of proteins."""
        return all([a.parent.is_residue() for a in self.atoms])

    def is_nucleotide(self):
        """Return True if all atoms in the atom group belong to nucleotides."""
        return all([a.parent.is_nucleotide() for a in self.atoms])

    def is_mixed(self):
        """Return True if the atoms in the atom group belong to different
        compound classes (water, hetero group, residue, or nucleotide)."""
        return len(set([a.parent.get_class() for a in self.atoms])) > 1

    def has_water(self):
        """Return True if at least one atom in the atom group belongs to a
        water molecule."""
        return any([a.parent.is_water() for a in self.atoms])

    def has_hetatm(self):
        """Return True if at least one atom in the atom group belongs to a
        hetero group, i.e., non-standard residues of proteins, DNAs, or RNAs,
        as well as atoms in other kinds of groups, such as carbohydrates,
        substrates, ligands, solvent, and metal ions."""
        return any([a.parent.is_hetatm() for a in self.atoms])

    def has_metal(self):
        """Return True if at least one atom in the atom group is a metal."""
        return any([a.parent.is_metal() for a in self.atoms])

    def has_residue(self):
        """Return True if at least one atom in the atom group belongs to a
        standard residue of proteins."""
        return any([a.parent.is_residue() for a in self.atoms])

    def has_nucleotide(self):
        """Return True if at least one atom in the atom group belongs to a
        nucleotide."""
        return any([a.parent.is_nucleotide() for a in self.atoms])

    def has_reference(self):
        """Return True if at least one compound is the reference of LUNA's
        analysis"""
        return any([a.parent.is_reference() for a in self.atoms])

    def as_json(self):
        """Represent the atom group as a dict containing the atoms, compounds,
        features, and compound classes (water, hetero group, residue,
        or nucleotide).

        The dict is defined as follows:

            * ``atoms`` (iterable of :class:`~luna.mol.atom.ExtendedAtom`): \
                    the list of atoms comprising the atom group;
            * ``compounds`` (iterable of \
                    :class:`~luna.pdb.core.residue.Residue`): the list of \
                    unique compounds that contain the atoms;
            * ``classes`` (iterable of str): the list of compound classes;
            * ``features`` (iterable of \
                    :class:`~luna.mol.features.ChemicalFeature`): the atom \
                    group's list of chemical features.
        """
        grp_obj = {}
        grp_obj["atoms"] = [atm.as_json() for atm in self.atoms]
        grp_obj["compounds"] = [comp.as_json() for comp in self.compounds]
        grp_obj["features"] = [feat.name for feat in self.features
                               if feat.name != "Atom"]
        grp_obj["classes"] = [comps.get_class() for comps in self.compounds]

        return grp_obj

    def clear_refs(self):
        """References to this `AtomGroup` instance will be removed from the
        list of atom groups of each atom in ``atoms``."""
        if self._recursive:
            for atm in self.atoms:
                atm.remove_atm_grps([self])

    def __repr__(self):
        return '<AtomGroup: [%s]>' % ', '.join([str(x) for x in self.atoms])

    def __eq__(self, other):
        """Overrides the default implementation"""
        if type(self) == type(other):
            return (self.atoms == other.atoms
                    and self.features == other.features)
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __lt__(self, other):
        atms1 = tuple(sorted(self.atoms))
        atms2 = tuple(sorted(other.atoms))
        return atms1 < atms2

    def __len__(self):
        # Number of atoms.
        return self.size

    def __hash__(self):
        """Overrides the default implementation"""
        if self._hash_cache is None:
            # Transform atoms and features list into an imutable data
            # structure. The lists are sorted in order to avoid
            # dependence on appending order.
            atoms_tuple = tuple(self.atoms)
            feat_tuple = tuple(self.features)
            self._hash_cache = hash((atoms_tuple, feat_tuple, self.__class__))
        return self._hash_cache


class PseudoAtomGroup(AtomGroup):
    """Represent only the atoms from an `AtomGroup` object that are involved
    in an interaction.

    Currently, this class is only used during the generation of LUNA's
    fingerprints.

    Parameters
    ----------
    parent_grp : `AtomGroup`
        The atom group that contains the subset of atoms ``atoms``.
    atoms : iterable of :class:`~luna.mol.atom.ExtendedAtom`
        A sequence of atoms.
    features : iterable of :class:`~luna.mol.features.ChemicalFeature`, \
                optional
        A sequence of chemical features.
    interactions : iterable of \
                :class:`~luna.interaction.type.InteractionType`, optional
        A sequence of interactions established by an atom group.
    """

    def __init__(self, parent_grp, atoms, features=None, interactions=None):
        self.parent_grp = parent_grp

        super().__init__(atoms, features, interactions, recursive=False)

    def __repr__(self):
        return ('<PseudoAtomGroup: [%s]>'
                % ', '.join([str(x) for x in self.atoms]))


class AtomGroupPerceiver():

    """Perceive and create atom groups for molecules.

    Parameters
    ----------
    feature_extractor : :class:`~luna.mol.features.FeatureExtractor`
        Perceive pharmacophoric properties from molecules.
    add_h : bool
        If True, add hydrogen to the molecules.
    ph : float, optional
        If not None, add hydrogens appropriate for pH ``ph``.
    amend_mol : bool
        If True, apply validation and standardization of molecules read
        from a PDB file.
    charge_model : class:`~luna.mol.charge_model.ChargeModel`
        A charge model object. By default, the implementation of OpenEye
        charge model is used.
    expand_selection : bool
        If True (the default), perceive features for a given molecule
        considering all nearby molecules. The goal is to identify any
        covalently bonded molecules that may alter the pharmacophoric
        properties or chemical functional groups.

        For instance, consider an amide of a peptide bond.
        If ``expand_selection`` is False, the residues forming the peptide
        bond will be analyzed separately, which will make the oxygen and the
        nitrogen of the amide to be perceived as carbonyl oxygen and amine,
        respectively. On the other hand, if ``expand_selection`` is True, the
        covalent bond between the residues will be identified and the amide
        will be correctly perceived.
    radius : float
        If ``expand_selection`` is True, select all molecules up to a maximum
        of ``radius`` away (measured in Å). The default value is 2.2, which
        comprises covalent bond distances.
    tmp_path : str, optional
        A temporary directory to where temporary files will be saved.
        If not provided, the system's default temporary directory will be used
        instead.
    critical : bool
        If False, ignore any errors during the processing a molecule and
        continue to the next one. The default value is True, which implies
        that any errors will raise an exception.
    """

    def __init__(self, feature_extractor, add_h=False, ph=None, amend_mol=True,
                 charge_model=OpenEyeModel(), expand_selection=True,
                 radius=COV_SEARCH_RADIUS, cache=None, tmp_path=None,
                 critical=True):

        self.feature_extractor = feature_extractor

        self.add_h = add_h
        self.ph = ph
        # If the user decided not to add hydrogens,
        # it will try to use the existing ones.
        self.keep_hydrog = not self.add_h

        self.amend_mol = amend_mol
        self.charge_model = charge_model
        self.expand_selection = expand_selection
        self.radius = radius

        self.cache = cache
        self.tmp_path = tmp_path

        # If the pharmacophoric perception is critical, any exception during
        # the processing of a molecule will stop the whole processing.
        self.critical = critical

    def perceive_atom_groups(self, compounds, mol_objs_dict=None):
        """Perceive and create atom groups for each molecule in ``compounds``.

        Parameters
        ----------
        compounds : iterable of :class:`~luna.pdb.core.residue.Residue`
            A sequence of molecules.
        mol_objs_dict : dict
            Map a compound, represented by its id, to a molecular object
            (:class:`~luna.wrappers.base.MolWrapper`,
            :class:`rdkit.Chem.rdchem.Mol`,
            or :class:`openbabel.pybel.Molecule`).

            This parameter can be used in cases where the ligand is read from a
            molecular file and no standardization or validation is required.

        Returns
        -------
         : `AtomGroupsManager`
            An `AtomGroupsManager` object containing all atom groups perceived
            for the molecules in ``compounds``.
        """
        mol_objs_dict = mol_objs_dict or {}

        self.atm_grps_mngr = AtomGroupsManager()
        self.atm_mapping = {}

        compounds = set(compounds)
        init_comps_set = set(compounds)

        if self.cache:
            cached_compounds = set([c for c in compounds
                                    if self.cache.is_compound_cached(c)])
            compounds = compounds - cached_compounds

        # Controls which compounds are allowed to expand
        # when 'expand_selection' is set ON.
        pruned_comps = set()
        # Create a queue of compounds to be processed.
        comp_queue = set(compounds)

        # Map Residue objects to a MolWrapper object.
        new_mol_objs_dict = {}
        # Initial compounds + border compounds.
        target_compounds = set()
        # Metals are prepared separatelly.
        metals = set()

        while comp_queue:
            comp = comp_queue.pop()

            # Skip molecules provided as a molecular object.
            if comp.id in mol_objs_dict:
                new_mol_objs_dict[comp] = mol_objs_dict[comp.id]
                continue

            # Add this compound to the binding site set.
            if not comp.is_metal():
                target_compounds.add(comp)
            else:
                metals.add(comp)

            if self.expand_selection:
                # Remove the ligand from the list when
                # it was provided as an OBMol object.
                comp_list = [c for c in get_proximal_compounds(comp)
                             if c.id not in mol_objs_dict]

                for prox_comp in comp_list:
                    if not prox_comp.is_metal():
                        target_compounds.add(prox_comp)
                    else:
                        metals.add(prox_comp)

                # Expands the queue of compounds
                # with any new border compound.
                if comp not in pruned_comps:
                    border_comps = set(comp_list) - compounds
                    pruned_comps |= border_comps
                    comp_queue |= border_comps

            # Otherwise, the compound list will be composed
            # only by the target compound.
            else:
                comp_list = [comp]

        # Stores atoms involved in metal coordination as a dict of dict,
        # where the first key is the residue and the second key is an atom.
        # The values are sets of metals (Residue objects).
        metals_coord = self._find_metal_coordination(target_compounds, metals)

        # Assign properties for ligands provided as external MOL files.
        for comp, mol_obj in new_mol_objs_dict.items():
            target_atoms = self._get_atoms(comp)
            self._assign_properties(mol_obj, target_atoms)

        # Assign properties for molecules from PDB files.
        if target_compounds:
            mol_obj, target_atoms = \
                self._get_mol_from_entity(target_compounds,
                                          metals_coord=metals_coord)
            self._assign_properties(mol_obj, target_atoms)

        # Assign properties to metals.
        if metals:
            self._assing_metal_properties(metals)

        # Apply cache.
        if self.cache and cached_compounds:
            print("\n\nSetting cache...\n\n")
            self._apply_cache(cached_compounds, comp.get_parent_by_level("M"))

        # Remove atom groups not comprising the provided
        # compound list (parameter 'compounds').
        remove_atm_grps = []
        for atm_grp in self.atm_grps_mngr:
            if any([c in init_comps_set for c in atm_grp.compounds]) is False:
                remove_atm_grps.append(atm_grp)
        self.atm_grps_mngr.remove_atm_grps(remove_atm_grps)

        return self.atm_grps_mngr

    def _find_metal_coordination(self, compounds, metals):

        # Stores atoms involved in metal coordination as a dict of dict,
        # where the first key is the residue and the second key is an atom.
        # The values are sets of metals (Residue objects).
        custom_dict = lambda: {"atm_idx": None, "metals": set()}
        metals_coord = defaultdict(lambda: defaultdict(custom_dict))

        for metal in metals:
            # Identifies potential dative bonds with metals.
            atm_pairs = get_contacts_with(metal, radius=METAL_COMPLEX_DIST)
            for atm1, atm2 in atm_pairs:
                if atm1.parent.is_metal() and not atm2.parent.is_metal():
                    other_atm = atm2
                elif not atm1.parent.is_metal() and atm2.parent.is_metal():
                    other_atm = atm1
                # Skip if the pair contains two metals or no metal.
                else:
                    continue

                # Skip pairs whose non-metal compound is not
                # in ``target_compounds``
                if other_atm.parent not in compounds:
                    continue

                # Skip pairs whose atom is not an O, N, or S.
                if other_atm.element not in ["O", "N", "S"]:
                    continue

                # Add this metal to the set of metals being
                # coordinated by the non-metal atom.
                other_atm.metal_coordination.add(metal)

                # Add the pair to the dict 'metals_coord'.
                metals_coord[other_atm.parent][other_atm]["metals"].add(metal)

        return metals_coord

    def _assign_properties(self, mol_obj, target_atoms=None):
        try:
            # Create a new MolWrapper object.
            mol_obj = MolWrapper(mol_obj)

            if mol_obj.get_num_heavy_atoms() != len(target_atoms):
                raise MoleculeSizeError("The number of heavy atoms in the PDB "
                                        "selection and in the MOL file are "
                                        "different.")

            # Ignore hydrogen atoms.
            atm_obj_list = [atm for atm in mol_obj.get_atoms()
                            if atm.get_atomic_num() != 1]

            atm_map = {}
            trgt_atms = {}
            ob_atms_map = {}
            for i, atm_obj in enumerate(atm_obj_list):
                atm_key = target_atoms[i].get_full_id()

                atm_map[atm_obj.get_idx()] = atm_key
                ob_atms_map[atm_key] = atm_obj

                trgt_atms[atm_key] = self._new_extended_atom(target_atoms[i])
                # Update atomic invariants for new ExtendedAtoms created.
                trgt_atms[atm_key].invariants = \
                    atm_obj.get_atomic_invariants()

            # Set all neighbors, i.e., covalently bonded atoms.
            for bond_obj in mol_obj.get_bonds():
                bgn_atm_obj = bond_obj.get_begin_atom()
                end_atm_obj = bond_obj.get_end_atom()

                # At least one of the atoms must be a non-hydrogen atom.
                if (bgn_atm_obj.get_atomic_num() != 1
                        or end_atm_obj.get_atomic_num() != 1):

                    # If the atom 1 is not a hydrogen, add atom 2 to its
                    # neighbor list.
                    if bgn_atm_obj.get_atomic_num() != 1:
                        full_id = atm_map.get(end_atm_obj.get_idx())
                        coord = \
                            mol_obj.get_atom_coord_by_id(end_atm_obj.get_id())
                        atom_info = AtomData(end_atm_obj.get_atomic_num(),
                                             coord,
                                             bond_obj.get_bond_type(),
                                             full_id)

                        bgn_atm = atm_map[bgn_atm_obj.get_idx()]
                        trgt_atms[bgn_atm].add_nb_info([atom_info])

                    # If the atom 2 is not a hydrogen, add atom 1 to its
                    # neighbor list.
                    if end_atm_obj.get_atomic_num() != 1:
                        full_id = atm_map.get(bgn_atm_obj.get_idx())
                        coord = \
                            mol_obj.get_atom_coord_by_id(bgn_atm_obj.get_id())
                        atom_info = AtomData(bgn_atm_obj.get_atomic_num(),
                                             coord,
                                             bond_obj.get_bond_type(),
                                             full_id)
                        end_atm = atm_map[end_atm_obj.get_idx()]
                        trgt_atms[end_atm].add_nb_info([atom_info])

            # Perceive pharmacophoric properties and create AtomGroup objects.
            group_features = \
                self.feature_extractor.get_features_by_groups(mol_obj, atm_map)
            for key in group_features:
                grp_obj = group_features[key]
                atoms = [trgt_atms[i] for i in grp_obj["atm_ids"]]

                self.atm_grps_mngr.new_atm_grp(atoms, grp_obj["features"])

            self._fix_pharmacophoric_rules(ob_atms_map)

            # Update the graph in the AtomGroupsManager object
            # with the current network.
            for atm in trgt_atms.values():
                for nb_info in atm.neighbors_info:
                    if nb_info.full_id in trgt_atms:
                        pair = atm, trgt_atms[nb_info.full_id]
                        self.atm_grps_mngr.graph.add_edge(*pair, weight=1)

        except Exception:
            logger.debug("Features were not correctly perceived.")

            if self.critical:
                raise

    def _get_default_invariants(self, pdb_atm):
        # Metal properties.
        data = DEFAULT_RES_DATA.get(pdb_atm.parent.resname, {})
        props = data.get("atoms", {})

        invariants = None
        if pdb_atm.name in props:
            invariants = props[pdb_atm.name]["invariants"]
            invariants = [int(i) for i in invariants.split(",")]
        return invariants

    def _assing_metal_properties(self, metals):
        for metal in metals:
            atms = self._get_atoms(metal)

            if len(atms) == 0:
                logger.error("No atom has been found for residue %s."
                             % metal.hierarchy_name)
                continue

            if len(atms) > 1:
                logger.error("An unexpected number of atoms (%d) has been "
                             "found for residue %s, while 1 was expected."
                             % (len(atms), metal.hierarchy_name))
                continue

            # Create/recover an extended atom object
            # and set its atomic invariant.
            atm = self._new_extended_atom(atms[0])
            atm.invariants = self._get_default_invariants(atm)

            # Define a new group and add it to 'atm_grps_mngr'.
            feats = [ChemicalFeature("Atom"), ChemicalFeature("Metal")]
            self.atm_grps_mngr.new_atm_grp([atm], feats)

    def _fix_pharmacophoric_rules(self, atms_map):
        """ Fix atom groups where one of its atoms coordinate a metal
        if necessary.

        For example, an imidazole group may be perceived as a
        positive ionizable group due to the set of pharmacophoric rules
        even when its nitrogens are coordinating a metal. In this case,
        we should not consider the group as ionizable as the N's lone
        pairs are involved in the dative bond."""
        for atm_grp1 in self.atm_grps_mngr:
            atoms = atm_grp1.atoms

            if not any([atm.has_metal_coordination() for atm in atoms]):
                continue

            is_imidazole = False
            for atm in atoms:
                if atm.element not in ["O", "N", "S"]:
                    continue

                invariants = atm.invariants
                atm_obj = atms_map[atm.get_full_id()]

                # Imidazole-like, but not tetrazoles.
                tetrazole_smarts = \
                    "$([nR1r5;$(n:n:n:n:c),$(n:n:n:c:n)])"
                imidazole_smarts = \
                    ("[$([n;H1]cn),$(nc[n;H1]),$([n;H0-1]cn);R1r5;"
                     f"!{tetrazole_smarts}]")
                is_imidazole = \
                    atm_obj.matches_smarts(imidazole_smarts)

                for atm_grp2 in atm.atm_grps:
                    # Skip groups containing more than one atom.
                    if len(atm_grp2.atoms) > 1:
                        continue

                    features = []
                    for feature in atm_grp2.features:
                        # If an atom is coordinating a metal
                        # and it is a donor, but it has no H
                        # bound to it, remove this feature.
                        if (atm.has_metal_coordination()
                                and feature.name == "Donor"
                                and invariants[5] == 0):
                            continue

                        # If a N is coordinating a metal
                        # and it is an acceptor, remove this
                        # feature because its lone pair is already
                        # being shared with the metal.
                        if (atm.has_metal_coordination()
                                and feature.name == "Acceptor"
                                and atm.element == "N"):
                            continue

                        if (atm.has_metal_coordination()
                                and feature.name == "PositivelyIonizable"
                                and atm.element == "N"):
                            continue

                        # If the imidazole N having an H bound to
                        # it (NH) is not coordinating a metal,
                        # obligatory the other N is doing so,
                        # otherwise this N would not reach this
                        # part of the code.
                        #
                        # In this case, remove its feature
                        # 'Acceptor' because the ring cannot form
                        # tautomers anymore.
                        if (not atm.has_metal_coordination()
                                and feature.name == "Acceptor"
                                and atm.element == "N"
                                and is_imidazole):
                            continue

                        features.append(feature)

                    # Update this atom group's features.
                    atm_grp2.features = features

            if is_imidazole:
                features = []
                for feature in atm_grp1.features:
                    if feature.name == "PositivelyIonizable":
                        continue
                    features.append(feature)
                atm_grp1.features = features

    def _apply_cache(self, compounds, model):

        def copy_atom(ref_atm, force_update=False):
            ref_res = ref_atm.parent
            ref_chain = ref_res.parent

            # Recover current structure entities.
            chain = model[ref_chain.id]
            res = chain[ref_res.id]
            atm = res[ref_atm.id]

            # If an ExtendedAtom already exists, return it.
            if not force_update and atm in self.atm_mapping:
                return self.atm_mapping[atm]

            # Define a new ExtendedAtom.
            ext_atm = self._new_extended_atom(atm)

            # Set its atomic invariants.
            ext_atm.invariants = ref_atm.invariants

            # Add neighbors' information to the new ExtendedAtom.
            ext_atm.neighbors_info = []
            for nbi in ref_atm.neighbors_info:
                atom_info = AtomData(nbi.atomic_num,
                                     nbi.coord, nbi.bond_type,
                                     nbi.full_id)
                ext_atm.add_nb_info([atom_info])

            return ext_atm

        comp_names = set([c.hierarchy_name for c in compounds])

        for atm_grp in sorted(self.cache.atm_grps_mngr):
            if any([c.hierarchy_name in comp_names for c in atm_grp.compounds]):
                atoms = []
                for atm in atm_grp.atoms:
                    # Copy extended atom, i.e., create a new extended
                    # atom or recover an existing one and set invariants
                    # and neighbors' information when necessary.
                    ext_atm = copy_atom(atm, force_update=True)
                    atoms.append(ext_atm)

                # Current features.
                features = atm_grp.features

                # Create the new atom group.
                atm_grp = self.atm_grps_mngr.new_atm_grp(atoms)
                atm_grp.features = features

        for edge in sorted(self.cache.atm_grps_mngr.graph.edges):
            # Skip already existing edges.
            if edge in self.atm_grps_mngr.graph.edges:
                continue

            atoms = []
            for atm in edge:
                # Copy extended atom, i.e., create a new
                # extended atom or recover an existing one
                # and set invariants and neighbors' information
                # if necessary.
                ext_atm = copy_atom(atm)
                atoms.append(ext_atm)

            # Update the AtomGroupsManager object with a new edge.
            self.atm_grps_mngr.graph.add_edge(atoms[0], atoms[1], weight=1)

    def _new_extended_atom(self, atm, invariants=None):
        if atm not in self.atm_mapping:
            self.atm_mapping[atm] = ExtendedAtom(atm, invariants=invariants)

        return self.atm_mapping[atm]

    def _get_atoms(self, compound):
        selector = Selector(keep_altloc=False,
                            keep_hydrog=self.keep_hydrog)

        return [atm for atm in compound.get_unpacked_list()
                if selector.accept_atom(atm)]

    def _get_mol_from_entity(self, compounds, metals_coord=None):

        atoms = [atm
                 for comp in sorted(compounds,
                                    key=lambda c: (c.parent.parent.id,
                                                   c.parent.id, c.idx))
                 for atm in self._get_atoms(comp)]

        model = atoms[0].get_parent_by_level('M')
        atom_selector = AtomSelector(atoms, keep_altloc=False,
                                     keep_hydrog=self.keep_hydrog)

        mol_obj, ignored_atoms = \
            biopython_entity_to_mol(model, atom_selector,
                                    amend_mol=self.amend_mol,
                                    add_h=self.add_h, ph=self.ph,
                                    metals_coord=metals_coord,
                                    tmp_path=self.tmp_path)

        # If the add_h property is set to False, the code will not remove any
        # existing hydrogens from the PDB structure. In these situations, the
        # list of atoms may contain hydrogens. But, we do not need to attribute
        # properties to hydrogens. We just need them to correctly set
        # properties to heavy atoms. So let's just ignore them.
        target_atoms = [atm for atm in atoms if atm.element != "H"
                        and atm not in ignored_atoms]

        return mol_obj, target_atoms



from Bio.PDB.Residue import Residue as BioResidue
from Bio.PDB.Residue import DisorderedResidue as BioDisorderedResidue
from Bio.PDB.Polypeptide import is_aa

from luna.pdb.core.entity import Entity


# List of metal residue names (based on PDB naming conventions)
METALS = ["LI", "NA", "K", "RB", "CS", "MG", "CA", "SR", "BA", "V", "CR",
          "MN", "MN3", "FE2", "FE", "CO", "3CO", "NI", "3NI", "CU1", "CU",
          "CU3", "ZN", "AL", "GA", "ZR", "MO", "4MO", "6MO", "RU", "RH",
          "RH3", "PD", "AG", "CD", "IN", "W", "RE", "OS", "OS4", "IR",
          "IR3", "PT", "PT4", "AU", "AU3", "HG", "TL", "PB", "BS3", "0BE",
          "4TI", "Y1", "YT3", "TA0"]


class Residue(BioResidue, Entity):
    """
    Custom Residue subclass with support for indexing, classification,
    hierarchical naming, and processing flags.

    Parameters
    ----------
    id : tuple
        Biopython residue ID (hetflag, resseq, icode).
    resname : str
        Residue name (e.g., "ALA", "HOH").
    segid : str
        Segment ID.
    idx : int, optional
        Index in parent chain.
    at_line : int, optional
        Line number in the original PDB.
    """

    def __init__(self, id, resname, segid, idx=None, at_line=None):
        super().__init__(id, resname, segid)

        # Store position within parent chain
        self.idx = idx
        # Optional PDB line number
        self.at_line = at_line 

        # Used by FTMapParser, optional.
        self.cluster_id = None
        
        # By default: no residue is a reference for processing.
        self._is_reference = False
    
    def __repr__(self):
        """Return the residue full id."""
        resname = self.get_resname()
        hetflag, resseq, icode = self.get_id()
        full_id = (resname, hetflag, resseq, icode)
        return "<Residue (LUNA subclass) %s het=%s resseq=%s icode=%s>" % full_id

    def __lt__(self, other):
        return self.idx < other.idx

    @property
    def metal_coordination(self):
        """
        Return atoms in this residue that participate in metal coordination.

        Returns
        -------
        dict
            Atom → set of coordination partners.
        """
        return {
            atom: atom.metal_coordination
            for atom in self.get_atoms()
            if getattr(atom, "has_metal_coordination", lambda: False)()
        }

    def is_reference(self):
        """
        Return True if this entity is marked as the reference for interaction analysis.

        This flag indicates that the user selected this entity
        as the *main subject* of interaction evaluation — interactions will be computed
        between this entity and others nearby (e.g., ligands, water, ions).

        Returns
        -------
        bool
            True if this entity is marked for interaction reference.
        """
        return self._is_reference

    def set_as_reference(self, is_reference=True):
        """
        Mark this entity as the reference for interaction analysis.

        This method sets a flag indicating that this residue should be
        treated as the main subject for calculating interactions 
        with nearby molecules.

        Parameters
        ----------
        is_reference : bool, optional
            If True, mark the entity as the interaction reference.
            If False, unmark it. Defaults to True.

        See Also
        --------
        is_reference : Check whether this entity is currently marked.
        """
        self._is_reference = is_reference

    def is_water(self):
        """Return True if the residue is a water molecule."""
        return self.get_id()[0] == "W"

    def is_metal(self):
        """Return True if the residue is a metal."""
        return (self.get_id()[0].startswith("H_")
                and self.resname in METALS)

    def is_hetatm(self):
        """Return True if the residue is an hetero group."""
        return (self.get_id()[0].startswith("H_")
                and not self.is_metal())

    def is_residue(self):
        """Return True if the residue is an amino acid."""
        return self.get_id()[0] == " " and is_aa(self.resname)

    def is_nucleotide(self):
        """Return True if the residue is a nucleic acid."""
        return self.get_id()[0] == " " and not self.is_residue()

    def is_monoatom(self):
        """Return True if the residue contains only one atom (or one non-hydrogen)."""
        atoms = list(self.get_atoms())
        return len(atoms) == 1 or len([a for a in atoms if getattr(a, "element", None) != "H"]) == 1

    def get_class(self):
        """Get the compound class."""
        if self.is_water():
            return "Water"
        if self.is_hetatm():
            return "Hetatm"
        if self.is_residue():
            return "Residue"
        if self.is_nucleotide():
            return "Nucleotide"
        return "Unknown"

    def get_flag(self):
        """
        Return a shorthand flag for residue classification.

        Returns
        -------
        str
            "H" for HETATM or metal, otherwise the original hetflag.
        """
        return "H" if self.get_id()[0].startswith("H_") else self.get_id()[0]

    def as_json(self):
        """
        Serialize this residue to a compact dictionary representation.

        Fields:
            chain, name, number, icode, repr

        'repr' will be 'sphere' for waters or monoatomic hetatms, 'stick' otherwise.
        """
        comp_repr = (
            "sphere" if self.is_water() or (self.is_hetatm() and self.is_monoatom()) else "stick"
        )
        return {
            "chain": self.get_parent().id,
            "name": self.resname,
            "number": self.id[1],
            "icode": self.id[2],
            "repr": comp_repr
        }


class DisorderedResidue(BioDisorderedResidue):
    """
    LUNA's DisorderedResidue placeholder.

    Currently identical to Biopython's, but allows future extensions and consistent imports.
    """
    def __init__(self, id):
        super().__init__(id)

    def __repr__(self):
        """Return disordered residue full identifier."""
        if self.child_dict:
            resname = self.get_resname()
            hetflag, resseq, icode = self.get_id()
            full_id = (resname, hetflag, resseq, icode)
            return "<DisorderedResidue (LUNA subclass) %s het=%s resseq=%i icode=%s>" % full_id
        else:
            return "<Empty DisorderedResidue (LUNA subclass)>"


import ast
from os import path
from pathlib import Path
import logging

from luna.util.config import Config
from luna.mol.entry import *
from luna.interaction.calc import InteractionCalculator
from luna.interaction.config import InteractionConfig
from luna.interaction.filter import InteractionFilter, BindingModeFilter
from luna.interaction.fp.type import IFPType
from luna.util.logging import VERBOSITY_LEVEL
from luna.util.exceptions import IllegalArgumentError


class ProjectParams(dict):

    """Define LUNA project parameters.

    Parameters
    ----------
    params : dict
        Set LUNA parameters from ``params``.

    fill_defaults : bool
        If True, initialize `ProjectParams` with default LUNA values.

        .. note::
            Any parameter provided in ``params`` will overwrite the
            default values.
    """

    def __init__(self, params=None, fill_defaults=False):

        params_to_parse = {}

        # Loads in default LUNA values.
        if fill_defaults:
            luna_path = path.abspath(path.join(path.realpath(__file__),
                                               '../..'))

            default_config_file = f"{luna_path}/config/default.cfg"
            config = Config(default_config_file)

            params_to_parse = config.as_dict(only_props=True)

            if params_to_parse["feat_cfg"] == "None":
                atom_prop_file = f"{luna_path}/data/LUNA.fdef"
                params_to_parse["feat_cfg"] = atom_prop_file

            if params_to_parse["inter_cfg"] == "None":
                inter_config_file = f"{luna_path}/interaction/config.cfg"
                params_to_parse["inter_cfg"] = inter_config_file

        # Overwrites default values with provided params.
        if params:
            params_to_parse.update(params)

        self._validate(params_to_parse)

    @classmethod
    def from_project_obj(cls, proj_obj):
        """Initialize a `ProjectParams` from a `~luna.projects.Project` object.

        Returns
        -------
         : `ProjectParams`
        """
        params = {"add_h": proj_obj.add_h,
                  "amend_mol": proj_obj.amend_mol,
                  "append_mode": proj_obj.append_mode,
                  "feat_cfg": proj_obj.atom_prop_file,
                  "binding_mode_filter": proj_obj.binding_mode_filter,
                  "ifp": proj_obj.calc_ifp,
                  "mfp": proj_obj.calc_mfp,
                  "entries": proj_obj.entries,
                  "ifp_count": proj_obj.ifp_count,
                  "ifp_diff_comp_classes": proj_obj.ifp_diff_comp_classes,
                  "ifp_length": proj_obj.ifp_length,
                  "ifp_num_levels": proj_obj.ifp_num_levels,
                  "ifp_output": proj_obj.ifp_output,
                  "ifp_radius_step": proj_obj.ifp_radius_step,
                  "ifp_sim_matrix_output": proj_obj.ifp_sim_matrix_output,
                  "ifp_type": proj_obj.ifp_type,
                  "inter_calc": proj_obj.inter_calc,
                  "logging_enabled": proj_obj.logging_enabled,
                  "mfp_output": proj_obj.mfp_output,
                  "nproc": proj_obj.nproc,
                  "pse": proj_obj.out_pse,
                  "overwrite_path": proj_obj.overwrite_path,
                  "pdb_path": proj_obj.pdb_path,
                  "ph": proj_obj.ph,
                  "pse_path": proj_obj.pse_path,
                  "use_cache": proj_obj.use_cache,
                  "verbosity": proj_obj.verbosity,
                  "working_path": proj_obj.working_path}

        return cls(params)

    def save_config_file(self, config_file):
        """Save the project parameters into a configuration file.

        Parameters
        ----------
        config_file : str
            The output configuration file.
        """
        with open(config_file, "w") as OUT:
            OUT.write("[entries]\n")
            for k, v in self._get_entry_params(config_file).items():
                OUT.write(f"{k} = {v}\n")
            OUT.write("\n")

            OUT.write("[paths]\n")
            OUT.write("working_path = %s\n" % self["working_path"])
            OUT.write("pdb_path = %s\n" % self["pdb_path"])
            OUT.write("overwrite_path = %s\n" % self["overwrite_path"])
            OUT.write("\n")

            OUT.write("[hydrogens]\n")
            OUT.write("add_h = %s\n" % self["add_h"])
            OUT.write("ph = %s\n" % self["ph"])
            OUT.write("\n")

            OUT.write("[standardization]\n")
            OUT.write("amend_mol = %s\n" % self["amend_mol"])
            OUT.write("\n")

            OUT.write("[features]\n")
            OUT.write("feat_cfg = %s\n" % self["atom_prop_file"])
            OUT.write("\n")

            OUT.write("[interactions]\n")
            for k, v in self._get_inter_params(config_file).items():
                OUT.write(f"{k} = {v}\n")
            OUT.write("\n")

            OUT.write("[binding_mode_filter]\n")
            for k, v in self._get_binding_filter_params(config_file).items():
                OUT.write(f"{k} = {v}\n")
            OUT.write("\n")

            OUT.write("[mfp]\n")
            OUT.write("mfp = %s\n" % self["calc_mfp"])
            OUT.write("mfp_output = %s\n" % self["mfp_output"])
            OUT.write("\n")

            OUT.write("[ifp]\n")
            OUT.write("ifp = %s\n" % self["calc_ifp"])
            OUT.write("ifp_num_levels = %s\n" % self["ifp_num_levels"])
            OUT.write("ifp_radius_step = %s\n" % self["ifp_radius_step"])
            OUT.write("ifp_length = %s\n" % self["ifp_length"])
            OUT.write("ifp_count = %s\n" % self["ifp_count"])
            OUT.write("ifp_type = %s\n" % self["ifp_type"].name)
            OUT.write("ifp_diff_comp_classes = %s\n"
                      % self["ifp_diff_comp_classes"])
            OUT.write("ifp_output = %s\n" % self["ifp_output"])
            OUT.write("ifp_sim_matrix_output = %s\n"
                      % self["ifp_sim_matrix_output"])
            OUT.write("\n")

            OUT.write("[pse]\n")
            OUT.write("pse = %s\n" % self["out_pse"])
            OUT.write("pse_path = %s\n" % self["pse_path"])
            OUT.write("\n")

            verbosity = [k for k, v in VERBOSITY_LEVEL.items()
                         if v == self["verbosity"]].pop()

            OUT.write("[general]\n")
            OUT.write("append_mode = %s\n" % self["append_mode"])
            OUT.write("use_cache = %s\n" % self["use_cache"])
            OUT.write("verbosity = %s\n" % verbosity)
            OUT.write("logging_enabled = %s\n" % self["logging_enabled"])
            OUT.write("nproc = %s\n" % self["nproc"])
            OUT.write("\n")

    def _get_entry_params(self, config_file):
        sep = ":"
        entries = []
        for e in self["entries"]:
            if isinstance(e, MolEntry):
                entries.append(e.to_string(sep))

            elif isinstance(e, MolFileEntry):
                fields = [str(e.pdb_id), str(e.mol_id), str(e.mol_file),
                          str(e.is_multimol_file)]
                entries.append(",".join(fields))

        path = Path(config_file).parent.resolve()
        filename = Path(config_file).stem

        entry_file = f"{path}/{filename}.entries.txt"
        with open(entry_file, "w") as OUT:
            for e in entries:
                OUT.write(f"{e}\n")

        return {"entries": entry_file,
                "pdb_id": None,
                "mol_file": None,
                "entries_sep": ":",
                "fields_sep": ","}

    def _get_inter_params(self, config_file):
        path = Path(config_file).parent.resolve()
        filename = Path(config_file).stem

        ic = self["inter_calc"]

        inter_config = ic.inter_config
        inter_cfg_file = None
        if isinstance(inter_config, InteractionConfig):
            inter_cfg_file = f"{path}/{filename}.inter.txt"
            inter_config.save_config_file(inter_cfg_file)

        inter_filter = ic.inter_filter
        filter_cfg_file = None
        if isinstance(inter_filter, InteractionFilter):
            filter_cfg_file = f"{path}/{filename}.filter.txt"
            inter_filter.save_config_file(filter_cfg_file)

        inter_filter = ic.inter_filter

        return {"inter_cfg": inter_cfg_file,
                "filter_cfg": filter_cfg_file,
                "add_non_cov": ic.add_non_cov,
                "add_cov": ic.add_cov,
                "add_proximal": ic.add_proximal,
                "add_atom_atom": ic.add_atom_atom,
                "add_dependent_inter": ic.add_dependent_inter,
                "add_h2o_pairs_with_no_ref":
                    ic.add_h2o_pairs_with_no_ref,
                "strict_donor_rules": ic.strict_donor_rules,
                "strict_weak_donor_rules": ic.strict_weak_donor_rules,
                "lazy_comps_list": ",".join(ic.lazy_comps_list)}

    def _get_binding_filter_params(self, config_file):
        path = Path(config_file).parent.resolve()
        filename = Path(config_file).stem

        bmf = self["binding_mode_filter"]
        bmf_cfg_file = None
        if isinstance(bmf, BindingModeFilter):
            bmf_cfg_file = f"{path}/{filename}.bind.txt"
            bmf.save_config_file(bmf_cfg_file)

        return {"bind_cfg": bmf_cfg_file}

    def _get_value(self, params, param_name, dtype,
                   fallback=None, is_none_valid=True):

        value = params.get(param_name, fallback)

        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            pass

        if is_none_valid and value is None:
            return value

        if type(value) != dtype:
            # Ok, if it's an int and the expected value is float.
            if dtype == float and type(value) == int:
                return value
            # Ok, if it's an integer represented as float.
            if (dtype == int and type(value) == float
                    and value.is_integer()):
                return int(value)

            raise TypeError("The parameter '%s' must be a(n) %s."
                            % (param_name, dtype.__name__))

        return value

    def _parse_entries_params(self, params):
        try:
            entries_file = self._get_value(params, "entries", str)

            pdb_id = self._get_value(params, "pdb_id", str)
            mol_file = self._get_value(params, "mol_file", str)
            entries_sep = self._get_value(params, "entries_sep", str)
            fields_sep = self._get_value(params, "fields_sep", str)

            mol_obj_type = params["mol_obj_type"]

            entries = None
            if entries_file:
                entries = list(Entry.from_file(entries_file, pdb_id, mol_file,
                                               entries_sep, fields_sep,
                                               mol_obj_type=mol_obj_type))
        except IllegalArgumentError:
            raise
        except Exception:
            entries = self._get_value(params, "entries", list)

        return {"entries": entries}

    def _parse_paths_params(self, params):
        working_path = self._get_value(params, "working_path", str)
        pdb_path = self._get_value(params, "pdb_path", str)
        overwrite_path = self._get_value(params, "overwrite_path", bool)

        return {"working_path": working_path,
                "pdb_path": pdb_path,
                "overwrite_path": overwrite_path}

    def _parse_hydrogens_params(self, params):
        add_h = self._get_value(params, "add_h", bool)
        ph = self._get_value(params, "ph", float)

        return {"add_h": add_h, "ph": ph}

    def _parse_standardization_params(self, params):
        amend_mol = self._get_value(params, "amend_mol", bool)

        return {"amend_mol": amend_mol}

    def _parse_features_params(self, params):
        config_file = self._get_value(params, "feat_cfg", str)

        return {"atom_prop_file": config_file}

    def _parse_filter_params(self, params):
        filter_types = ["pli", "ppi", "pni", "nni", "nli"]

        if "default_filter" in params:
            default_filter = params["default_filter"]

            if default_filter not in filter_types:
                raise IllegalArgumentError("The informed default filter '%s' "
                                           "is not valid. The valid default "
                                           "filters are: %s."
                                           % (default_filter,
                                              ", ".join(filter_types)))

            func_name = f"new_{default_filter}_filter"
            func = getattr(InteractionFilter, func_name)
            return func()

        else:
            filter_file = self._get_value(params, "filter_cfg", str)
            inter_filter = None
            if filter_file is not None:
                inter_filter = InteractionFilter.from_config_file(filter_file)

            return inter_filter

    def _parse_inter_params(self, params):

        try:
            ic = self._get_value(params, "inter_calc",
                                 InteractionCalculator,
                                 is_none_valid=False)

        except Exception:
            config_file = self._get_value(params, "inter_cfg", str)

            inter_config = None
            if config_file is not None:
                inter_config = InteractionConfig.from_config_file(config_file)

            inter_filter = self._parse_filter_params(params)

            add_non_cov = self._get_value(params, "add_non_cov", bool)

            add_cov = self._get_value(params, "add_cov", bool)

            add_proximal = self._get_value(params, "add_proximal", bool)

            add_atom_atom = self._get_value(params, "add_atom_atom", bool)

            add_dep_inter = self._get_value(params, "add_dependent_inter",
                                            bool)

            add_h2o_pairs = self._get_value(params,
                                            "add_h2o_pairs_with_no_ref",
                                            bool)

            sdonor_rules = self._get_value(params, "strict_donor_rules", bool)

            swdonor_rules = self._get_value(params, "strict_weak_donor_rules",
                                            bool)

            lazy_comps_list = self._get_value(params, "lazy_comps_list", str)

            if lazy_comps_list is not None:
                lazy_comps_list = lazy_comps_list.split(",")

            ic = None
            if inter_config is not None:
                params = {
                    "inter_config": inter_config,
                    "inter_filter": inter_filter,
                    "add_non_cov": add_non_cov,
                    "add_cov": add_cov,
                    "add_proximal": add_proximal,
                    "add_atom_atom": add_atom_atom,
                    "add_dependent_inter": add_dep_inter,
                    "add_h2o_pairs_with_no_ref": add_h2o_pairs,
                    "strict_donor_rules": sdonor_rules,
                    "strict_weak_donor_rules": swdonor_rules,
                    "lazy_comps_list": lazy_comps_list,
                }

                ic = InteractionCalculator(**params)

        return {"inter_calc": ic}

    def _parse_binding_mode_filter(self, params):
        try:
            bmf = self._get_value(params, "binding_mode_filter",
                                  BindingModeFilter,
                                  is_none_valid=False)
        except Exception:
            config_file = self._get_value(params, "bind_cfg", str)

            bmf = None
            if config_file is not None:
                bmf = BindingModeFilter.from_config_file(config_file)

        return {"binding_mode_filter": bmf}

    def _parse_mfp_params(self, params):
        calc_mfp = self._get_value(params, "mfp", bool)
        mfp_output = self._get_value(params, "mfp_output", str)

        return {"calc_mfp": calc_mfp, "mfp_output": mfp_output}

    def _parse_ifp_params(self, params):

        calc_ifp = self._get_value(params, "ifp", bool)

        num_levels = self._get_value(params, "ifp_num_levels", int)

        radius_step = self._get_value(params, "ifp_radius_step", float)

        ifp_length = self._get_value(params, "ifp_length", int)

        ifp_count = self._get_value(params, "ifp_count", bool)

        ifp_diff_comp_classes = self._get_value(params,
                                                "ifp_diff_comp_classes",
                                                bool)

        try:
            ifp_type = self._get_value(params, "ifp_type", str)

            if ifp_type is not None:
                try:
                    ifp_type = IFPType[ifp_type]
                except KeyError:
                    valid_ifps = ", ".join([e.name for e in IFPType])
                    raise KeyError("The accepted IFP types are: "
                                   "%s." % valid_ifps)
        except Exception:
            ifp_type = self._get_value(params, "ifp_type", IFPType)

        ifp_output = self._get_value(params, "ifp_output", str)

        ifp_sim_matrix_output = self._get_value(params,
                                                "ifp_sim_matrix_output", str)

        return {
            "calc_ifp": calc_ifp,
            "ifp_num_levels": num_levels,
            "ifp_radius_step": radius_step,
            "ifp_length": ifp_length,
            "ifp_count": ifp_count,
            "ifp_diff_comp_classes": ifp_diff_comp_classes,
            "ifp_type": ifp_type,
            "ifp_output": ifp_output,
            "ifp_sim_matrix_output": ifp_sim_matrix_output,
        }

    def _parse_pse_params(self, params):
        out_pse = self._get_value(params, "pse", bool)
        pse_path = self._get_value(params, "pse_path", str)

        return {"out_pse": out_pse, "pse_path": pse_path}

    def _parse_general_params(self, params):

        append_mode = self._get_value(params, "append_mode", bool)

        use_cache = self._get_value(params, "use_cache", bool)

        verbosity = self._get_value(params, "verbosity", int)

        if verbosity is not None:
            try:
                verbosity = [verbosity for k, v in VERBOSITY_LEVEL.items()
                             if v == verbosity].pop()
            except Exception:
                if verbosity not in VERBOSITY_LEVEL:
                    lvls = sorted(VERBOSITY_LEVEL.items())
                    msg = ("The informed verbosity level '%s' is not valid. "
                           "The valid levels are: %s."
                           % (repr(verbosity),
                              ", ".join(["%d (%s)"
                                         % (k, logging.getLevelName(v))
                                         for k, v in lvls])))
                    raise IllegalArgumentError(msg)

        logging_enabled = self._get_value(params, "logging_enabled", bool)

        nproc = self._get_value(params, "nproc", int)

        return {
            "append_mode": append_mode,
            "use_cache": use_cache,
            "verbosity": verbosity,
            "logging_enabled": logging_enabled,
            "nproc": nproc,
        }

    def _validate(self, params):
        validated_params = {}

        # Entries
        validated_params.update(self._parse_entries_params(params))

        # Paths
        validated_params.update(self._parse_paths_params(params))

        # Hydrogens
        validated_params.update(self._parse_hydrogens_params(params))

        # Standardization
        validated_params.update(self._parse_standardization_params(params))

        # Features
        validated_params.update(self._parse_features_params(params))

        # Interactions
        validated_params.update(self._parse_inter_params(params))

        # Binding mode filter
        validated_params.update(self._parse_binding_mode_filter(params))

        # MFPs
        validated_params.update(self._parse_mfp_params(params))

        # IFPs
        validated_params.update(self._parse_ifp_params(params))

        # PSE
        validated_params.update(self._parse_pse_params(params))

        # General
        validated_params.update(self._parse_general_params(params))

        super().__init__(validated_params)




from Bio.PDB.Entity import Entity as BioEntity


class Entity(BioEntity):
    """
    Base class for LUNA's Structure, Model, Chain, and Residue classes.

    Wraps child management, identifier logic, and hierarchical traversal.
    """

    @property
    def hierarchy_name(self):
        """
        Return a full hierarchical name for this entity.

        For example:
            - Structure:       "1ABC"
            - Model:           "1ABC/0"
            - Chain:           "1ABC/0/A"
            - Residue:         "1ABC/0/A/GLY/42"
            - Atom:            override in Atom class

        Returns
        -------
        str
            Hierarchical identifier.
        """
        full_id = self.get_full_id()

        if self.level == "S":
            return str(full_id[0])

        if self.level == "M":
            return f"{full_id[0]}/{full_id[1]}"

        if self.level == "C":
            return f"{full_id[0]}/{full_id[1]}/{full_id[2]}"

        if self.level == "R":
            parent_path = '/'.join(str(x) for x in full_id[0:3])
            resname = f"{self.resname}/{self.id[1]}{self.id[2].strip()}"
            return f"{parent_path}/{resname}"

        # Default fallback
        return "/".join(str(x) for x in full_id)

    def get_parent_by_level(self, level):
        """
        Traverse up the SMCRA hierarchy to retrieve a parent at a specified level.

        Parameters
        ----------
        level : str
            One of: "A", "R", "C", "M", "S"

        Returns
        -------
        Entity
            The parent entity at the requested level.

        Raises
        ------
        ValueError
            If the level is invalid or not higher than the current level.
        """
        levels = ("A", "R", "C", "M", "S")

        if level not in levels:
            raise ValueError(f"Level must be one of {levels}.")
        if levels.index(level) < levels.index(self.level):
            raise ValueError(
                f"Cannot traverse up to '{level}' from level '{self.level}'. "
                "Valid hierarchy: A < R < C < M < S."
            )
        if self.level == level:
            return self
        return self.parent.get_parent_by_level(level)


import numpy as np
from itertools import product
from collections import defaultdict

from rdkit.Chem import SanitizeFlags, SanitizeMol
from openbabel.openbabel import OBSmartsPattern

from luna.wrappers.base import BondType, AtomWrapper, MolWrapper
from luna.mol.charge_model import OpenEyeModel
import luna.util.math as lm

import logging

logger = logging.getLogger()


class MolValidator:
    """Validate and fix molecules with the errors most commonly
    found when parsing PDB files with Open Babel.

    Parameters
    ----------
    charge_model : :class:`~luna.mol.charge_model.ChargeModel`
        A charge model object. By default, the implementation of OpenEye charge
        model is used.
    fix_nitro : bool
        If True, fix nitro groups whose nitrogens are perceived as having a
        valence of 5 and that are double-bonded to the oxygens.
        SMILES representation of the error: '[$([NX3v5]([!#8])(=O)=O)]'.
    fix_amidine_and_guanidine : bool
        If True, fix amidine and guanidine-like groups.
        When the molecule is ionized, it may happen that the charge is
        incorrectly assigned to the central carbon, which ends up with a +1
        charge and the nitrogen double-bonded to it ends up with a +0 charge.
    fix_valence : bool
        If True, fix the valence of atoms.
        Currently, we only detect and fix errors of quaternary ammonium
        nitrogens. These charged atoms are sometimes perceived as having a
        valence equal to 5 and no charge, i.e., Open Babel considers those
        nitrogens as hypervalent.
    fix_charges : bool
        If True, fix charges on basis of the charge model ``charge_model``.
    """

    def __init__(self,
                 charge_model=OpenEyeModel(),
                 fix_nitro=True,
                 fix_amidine_and_guanidine=True,
                 fix_valence=True,
                 fix_charges=True,
                 fix_in_ring=True,
                 metals_coord=None):
        self.charge_model = charge_model
        self.fix_nitro = fix_nitro
        self.fix_amidine_and_guanidine = fix_amidine_and_guanidine
        self.fix_valence = fix_valence
        self.fix_charges = fix_charges
        self.fix_in_ring = fix_in_ring
        self.metals_coord = metals_coord or []

    def validate_mol(self, mol_obj, pdb_mapping):
        """Validate molecule ``mol_obj``.

        It will first try to fix the provided molecule according to the
        initialization flags. If any errors are detected and if it fails
        to fix them, then this function may return False, i.e., the molecule
        is invalid.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, \
                :class:`rdkit.Chem.rdchem.Mol`, \
                or :class:`openbabel.pybel.Molecule`
            The molecule.

        Returns
        -------
        is_valid : bool
            If the molecule is valid or not.
        """

        if not isinstance(mol_obj, MolWrapper):
            mol_obj = MolWrapper(mol_obj)

        # Validations to be made...
        #       TODO: check aromaticity.
        #       TODO: return list of errors

        if self.metals_coord:
            self._fix_atoms_bound_to_metals(mol_obj)

        if self.fix_nitro:
            self._fix_nitro_substructure_and_charge(mol_obj)

        if self.fix_amidine_and_guanidine:
            self._fix_amidine_and_guanidine_charges(mol_obj)

        # Check if the molecule has errors...
        is_valid = True
        for atm_obj in mol_obj.get_atoms():

            pdb_atm = (pdb_mapping[atm_obj.get_idx()]
                       if atm_obj.get_idx() in pdb_mapping
                       else None)

            if not self._is_in_ring_valid(atm_obj, pdb_atm):
                is_valid = False

            if not self._is_valence_valid(atm_obj):
                is_valid = False

            if not self._is_charge_valid(atm_obj):
                is_valid = False

        if not is_valid:
            logger.debug("Invalid molecule: check the logs for more "
                         "information.")

        return is_valid

    def _fix_atoms_bound_to_metals(self, mol_obj):

        amine_smarts = "[$([NX4+;H3,H2,H1;!$(NC=O)][C])]"

        # Imidazole-like, but not tetrazoles.
        imidazole_smarts = ("[$(nc[n;H1]);"
                            "!$([nR1r5;$(n:n:n:n:c),"
                            "$(n:n:n:c:n)])]")

        # Iterate over each residue and that has a dative bond
        # with a metal to evaluate if its charge is incorrect.
        for res in self.metals_coord:
            for pdb_atm, data in self.metals_coord[res].items():

                if data["atm_idx"] is None:
                    continue

                atm_obj = mol_obj.get_atom_by_idx(data["atm_idx"])
                metals = data["metals"]

                if not atm_obj:
                    raise KeyError("Atom #%d does not exist."
                                   % data["atm_idx"])

                # Fix HIS:ND1 and HIS:NE2 that are bound to metals
                #   and other imidazole groups.
                #
                # It only fixes imidazole Ns having a +1 charge, i.e, the
                # N with the double bond. For these Ns, Open Babel tend to
                # protonate the N even if it has being neutralized by the
                # standardization method.
                if ((res.resname == "HIS"
                        and pdb_atm.name in ["ND1", "NE2"])
                        or (atm_obj.matches_smarts(imidazole_smarts))):

                    if atm_obj.get_charge() == 1:
                        logger.debug("Atom #%d has incorrect charge. It will "
                                     "update its charge from 1 to 0."
                                     % atm_obj.get_idx())

                        atm_obj.set_charge(0)
                        self._remove_explicit_hydrogens(atm_obj)

                # Fix main chain N or LYS:NZ that are bound to metals and
                # that has a charge = +1.
                elif ((res.resname == "LYS" and pdb_atm.name == "NZ")
                        or pdb_atm.name == "N"
                        or (pdb_atm.parent.is_hetatm()
                            and atm_obj.get_symbol() == "N"
                            and atm_obj.matches_smarts(amine_smarts))):

                    if atm_obj.get_charge() == 1:

                        logger.debug("Atom #%d has incorrect charge. It will "
                                     "update its charge from 1 to 0 and fix "
                                     "its number of bound hydrogens."
                                     % atm_obj.get_idx())

                        # Set residue charge to 0.
                        atm_obj.set_charge(0)

                        # Now, loop over this atom's bonds to identify
                        # all bonds with hydrogens. Then, it will identify
                        # the closest H to the metal and remove it.
                        #
                        # That's because residue can only coordinate
                        # one metal at once.
                        h_to_remove = None
                        min_dist = float('inf')
                        for b in atm_obj.get_bonds():
                            if b.get_partner_atom(atm_obj).get_symbol() == "H":
                                h_obj = b.get_partner_atom(atm_obj)

                                for metal in metals:
                                    metal_atm = list(metal.get_atoms())[0]

                                    c1 = h_obj.get_coord()
                                    c2 = metal_atm.coord
                                    dist = lm.euclidean_distance(c1, c2)
                                    # Distance between the H and the metal.
                                    if dist < min_dist:
                                        h_to_remove = h_obj
                                        min_dist = dist

                        # Remove the closest H to the metal.
                        if h_to_remove:
                            ob_mol = atm_obj.parent.unwrap()
                            ob_mol.DeleteAtom(h_to_remove.unwrap())

                # Fix ligands' phenol OH.
                elif res.is_hetatm() and atm_obj.matches_smarts("[OH+0]c"):
                    atm_obj.set_charge(-1)
                    self._remove_explicit_hydrogens(atm_obj)

    def _fix_nitro_substructure_and_charge(self, mol_obj):
        ob_smart = OBSmartsPattern()
        # Invalid nitro pattern.
        ob_smart.Init("[$([NX3v5]([!#8])(=O)=O)]")
        if ob_smart.Match(mol_obj.unwrap()):
            logger.debug("One or more invalid nitro substructures "
                         "('*-N(=O)=O') were found. It will try to "
                         "substitute them to '*-[N+]([O-])=O'.")

            # Iterate over each Nitro group in the molecule.
            for ids in ob_smart.GetUMapList():
                # Get the N atom.
                atm_obj = [AtomWrapper(mol_obj.GetAtom(i)) for i in ids][0]
                for bond in atm_obj.get_bonds():
                    partner = bond.get_partner_atom(atm_obj)

                    if (partner.get_symbol() == "O"
                            and bond.get_bond_type() == BondType.DOUBLE):
                        # Change double bond to single bond.
                        bond.set_bond_type(BondType.SINGLE)
                        # Attributes a +1 charge to the N.
                        atm_obj.set_charge(1)
                        # Attributes a -1 charge to the O.
                        partner.set_charge(-1)

                        # It needs to update only one of the oxygen bonds.
                        break

            ob_smart = OBSmartsPattern()
            # Valid nitro pattern.
            ob_smart.Init("[$([NX3v4+](=O)[O-])][!#8]")
            if ob_smart.Match(mol_obj.unwrap()):
                logger.debug("Invalid nitro substructures ('*-N(=O)=O') "
                             "successfully substituted to '*-[N+]([O-])=O'.")

    def _fix_amidine_and_guanidine_charges(self, mol_obj):
        # These errors occur with guanidine-like substructures when the
        # molecule is ionized. It happens that the charge is incorrectly
        # assigned to the central carbon, so the guanidine-like C ends up with
        # a +1 charge and the N with a double bond to the central C ends up
        # with a +0 charge. To fix it, we assign the correct charges to
        # the N (+1) and C (0).

        ob_smart = OBSmartsPattern()
        # Invalid amidine and guanidine pattern.
        ob_smart.Init("[$([NH1X2v3+0](=[CH0X3+1](N)))]")
        if ob_smart.Match(mol_obj.unwrap()):
            logger.debug("One or more amidine/guanidine substructures with no "
                         "charge were found. It will try to attribute a +1 "
                         "charge to the N bound to the central carbon with a "
                         "double bond.")

            # Iterate over each Amidine/Guanidine group in the molecule.
            for ids in ob_smart.GetUMapList():
                # Get the N atom.
                atm_obj = [AtomWrapper(mol_obj.GetAtom(i)) for i in ids][0]

                for bond in atm_obj.get_bonds():
                    partner = bond.get_partner_atom(atm_obj)

                    if partner.get_symbol() == "C":
                        # Attributes a +1 charge to the N and corrects its
                        # degree to three, i.e., the N will become '=NH2+'.
                        atm_obj.set_charge(1)

                        # Set the number of implicit Hydrogens to 1 (=NH2+).
                        #
                        # Explanation: the above initialized Smarts pattern
                        #      checks for nitrogens containing 1 explicit
                        #      hydrogen ([NH1X2v3+0]). Thus, to correctly
                        #      update its valence to 4 (=NH2+), it needs to
                        #      add a new implicit hydrogen.
                        atm_obj.unwrap().SetImplicitHCount(1)

                        # Remove any charges in the C.
                        partner.set_charge(0)

            ob_smart = OBSmartsPattern()
            # Valid amidine and guanidine pattern.
            ob_smart.Init("[$([NH2X3v4+1](=[CH0X3+0](N)))]")
            if ob_smart.Match(mol_obj.unwrap()):
                logger.debug("Invalid amidine/guanidine substructures were "
                             "correctly charged.")

    def _is_in_ring_valid(self, atm_obj, pdb_atm):

        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        if pdb_atm is None:
            return True

        in_ring = set([('HIS', 'CD2'), ('HIS', 'CE1'), ('HIS', 'CG'),
                       ('HIS', 'ND1'), ('HIS', 'NE2'), ('PHE', 'CD1'),
                       ('PHE', 'CD2'), ('PHE', 'CE1'), ('PHE', 'CE2'),
                       ('PHE', 'CG'), ('PHE', 'CZ'), ('PRO', 'CA'),
                       ('PRO', 'CB'), ('PRO', 'CD'), ('PRO', 'CG'),
                       ('PRO', 'N'), ('TRP', 'CD1'), ('TRP', 'CD2'),
                       ('TRP', 'CE2'), ('TRP', 'CE3'), ('TRP', 'CG'),
                       ('TRP', 'CH2'), ('TRP', 'CZ2'), ('TRP', 'CZ3'),
                       ('TRP', 'NE1'), ('TYR', 'CD1'), ('TYR', 'CD2'),
                       ('TYR', 'CE1'), ('TYR', 'CE2'), ('TYR', 'CG'),
                       ('TYR', 'CZ')])

        key = (pdb_atm.parent.resname, pdb_atm.name)

        # If Open Babel perceives the current atom as belonging to a ring,
        #   why it shouldn't, try to fix it.
        if (pdb_atm.parent.is_residue() and atm_obj.is_in_ring()
                and key not in in_ring):

            logger.debug("Atom #%d has been incorrectly perceived as part of "
                         "a ring." % atm_obj.get_idx())

            if self.fix_in_ring:
                logger.debug("'Fix in ring' option is set on. It will "
                             "set the atom #%d as not part of a ring."
                             % atm_obj.get_idx())

                atm_obj.set_in_ring(False)

                return True
            return False
        return True

    def _is_valence_valid(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        # Currently, atoms other than N are not evaluated as no valence error
        # has been identified for them.
        if atm_obj.get_atomic_num() == 7:

            # Molecules containing quaternary ammonium N.
            #
            #    - Open Babel 2.3.2 incorrectly sets N with:
            #         . Valence = 5
            #         . Charge  = 0
            #         . # Implicit Hs = 0
            #
            #    - Solution: set the number of implicit Hs to 0 and set the
            #                charge to +1.
            #
            if atm_obj.get_valence() == 5 and atm_obj.get_charge() == 0:
                logger.debug("Atom #%d has incorrect valence and charge."
                             % atm_obj.get_idx())

                if self.fix_valence:
                    logger.debug("'Fix valence' option is set on. It will "
                                 "update the valence of atom #%d from %d "
                                 "to 4 and correct its charge."
                                 % (atm_obj.get_idx(), atm_obj.get_valence()))

                    # Fix the number of implicit Hs and charge.
                    atm_obj.unwrap().SetImplicitHCount(0)
                    atm_obj.set_charge(1)
                    return True
                return False

            # Molecules containing quaternary ammonium N.
            #
            #    - Open Babel 3.x incorrectly sets N with:
            #         . Valence = 5
            #         . Charge  = 1
            #         . Num. Implicit Hs = 0
            #         . Num. Explicit Hs = 1
            #
            #    - Solution: remove the H and its bond with the N.
            #
            elif (atm_obj.get_valence() == 5 and atm_obj.get_charge() == 1
                    and atm_obj.get_h_count() == 1):

                logger.debug("Atom #%d has incorrect valence and number of "
                             "hydrogens." % atm_obj.get_idx())

                if self.fix_valence:
                    logger.debug("'Fix valence' option is set on. It will "
                                 "update the valence of atom #%d from %d "
                                 "to 4 and correct the number of hydrogens "
                                 "bound to it."
                                 % (atm_obj.get_idx(), atm_obj.get_valence()))

                    self._remove_explicit_hydrogens(atm_obj)

                    return True
                return False

            # Fix amide N with an incorrect assigned charge.
            #
            # E.g: 1BK0:A:FE:350
            # E.g: 1BLZ:A:FE:332
            elif atm_obj.matches_smarts("[$([#7;X4H2+1](C)C=O)]"):

                logger.debug("Atom #%d has incorrect valence, charge, and "
                             "number of hydrogens." % atm_obj.get_idx())

                if not self.fix_valence:
                    return False

                # Find all Cs and Hs bound to the N.
                C_atms = []
                H_atms = []
                for bond_obj in atm_obj.get_bonds():
                    partner_obj = bond_obj.get_partner_atom(atm_obj)

                    if partner_obj.get_symbol() == "H":
                        H_atms.append(partner_obj)
                    else:
                        C_atms.append(partner_obj)

                # N coordinate.
                N_coord = np.array(atm_obj.get_coord())
                # Cs coordinates.
                C_coords = [np.array(c.get_coord()) for c in C_atms]

                # CNC normal vector.
                CNC_normal = lm.calc_normal(C_coords + [N_coord])

                # Measures the displacement angle between the
                # normal vector and the NH vector.
                H_data = defaultdict(dict)
                for H_atm in H_atms:
                    NH_vect = (np.array(H_atm.get_coord())
                               - np.array(atm_obj.get_coord()))
                    disp_angle = lm.angle(CNC_normal, NH_vect)

                    H_data[H_atm.get_idx()]["atm_obj"] = H_atm
                    H_data[H_atm.get_idx()]["disp"] = disp_angle

                # Measures the angles between each pair of C, N, and H.
                for C_atm, H_atm in product(C_atms, H_atms):
                    NC_vect = (np.array(C_atm.get_coord()) - N_coord)
                    NH_vect = (np.array(H_atm.get_coord()) - N_coord)

                    key = ("angle1" if "angle1" not in H_data[H_atm.get_idx()]
                           else "angle2")

                    H_data[H_atm.get_idx()][key] = lm.angle(NH_vect, NC_vect)

                # Verifies if the angles formed by the Hs.
                H_to_remove = []
                for key in H_data:
                    angle_diff = abs(H_data[key]["angle1"]
                                     - H_data[key]["angle2"])
                    disp_diff = abs(H_data[key]["disp"] - 90)

                    # The angle between the Cs, N, and H should be nearly
                    # equal, while the H must be in the same plane as the
                    # C-N-C. If it does not match this criteria, remove
                    # the H.
                    if angle_diff > 1 or disp_diff > 1:
                        H_to_remove.append(H_data[key]["atm_obj"])

                # If there are no or more than two Hs to remove, return False.
                if len(H_to_remove) != 1:
                    return False

                logger.debug("'Fix valence' option is set on. It will "
                             "update the valence of atom #%d from %d "
                             "to 3 and correct its charge and the number "
                             "of hydrogens bound to it."
                             % (atm_obj.get_idx(), atm_obj.get_valence()))

                ob_mol = atm_obj.parent.unwrap()
                ob_mol.DeleteAtom(H_to_remove[0].unwrap())

                atm_obj.set_charge(0)

                return True
        return True

    def _is_charge_valid(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        expected_charge = self._get_expected_charge(atm_obj)

        if (expected_charge is not None
                and expected_charge != atm_obj.get_charge()):
            logger.debug("Atom #%d has incorrect charges defined."
                         % atm_obj.get_idx())

            if self.fix_charges:
                logger.debug("'Fix charges' option is set on. It will update "
                             "the charge of atom #%d from %d to %d."
                             % (atm_obj.get_idx(), atm_obj.get_charge(),
                                expected_charge))
                atm_obj.set_charge(expected_charge)
                return True
            return False
        return True

    def _get_expected_charge(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        return self.charge_model.get_charge(atm_obj)

    def _remove_explicit_hydrogens(self, atm_obj):
        if not isinstance(atm_obj, AtomWrapper):
            atm_obj = AtomWrapper(atm_obj)

        delete_hs = []
        for b in atm_obj.get_bonds():
            if b.get_partner_atom(atm_obj).get_symbol() == "H":
                delete_hs.append(b.get_partner_atom(atm_obj))

        for hs_obj in delete_hs:
            atm_obj.parent.unwrap().DeleteAtom(hs_obj.unwrap())


class RDKitValidator:
    """Check if RDKit molecular objects are valid or not.

    Parameters
    ----------
    sanitize_opts : :class:`rdkit.Chem.rdchem.SanitizeFlags`
        Sanitization operations to be carried out.
    """

    def __init__(self, sanitize_opts=SanitizeFlags.SANITIZE_ALL):
        self.sanitize_opts = sanitize_opts

    def is_valid(self, rdk_mol):
        """Try to sanitize the molecule ``rdk_mol``.
        If it succeeds, returns True. Otherwise, returns False.

        Parameters
        ----------
        rdk_mol : :class:`rdkit.Chem.rdchem.Mol`

        Returns
        -------
        is_valid : bool
            If the molecule is valid or not.
        """
        try:
            SanitizeMol(rdk_mol, sanitizeOps=self.sanitize_opts)
            return True
        except Exception as e:
            logger.exception(e)
            return False



import os
from pathlib import Path
from subprocess import Popen, PIPE

from luna.docking.dock6.params import Params
from luna.util.default_values import DOCK6
from luna.util.exceptions import ProcessingFailed


import logging
logger = logging.getLogger()


class Dock6Manager:

    def __init__(self, dock6=DOCK6):

        self.dock6 = dock6

    def _validate_params(self, params):

        if "ligand_atom_file" not in params:
            raise KeyError("The parameter 'ligand_atom_file' has not been "
                           "defined in the input file.")
        elif not os.path.exists(params["ligand_atom_file"]):
            raise FileNotFoundError("The ligand atom file defined in "
                                    "the input file was not found.")
        elif (" " in params["ligand_atom_file"]
                or "\t" in params["ligand_atom_file"]):
            raise ValueError("Spaces or tabs were found in the ligand "
                             "atom file. However, DOCK 6 cannot properly "
                             "deal with them.")

        if "receptor_site_file" not in params:
            raise KeyError("The parameter 'receptor_site_file' has not been "
                           "defined in the input file.")
        elif not os.path.exists(params["receptor_site_file"]):
            raise FileNotFoundError("The receptor site file defined in "
                                    "the input file was not found.")
        elif (" " in params["receptor_site_file"]
                or "\t" in params["receptor_site_file"]):
            raise ValueError("Spaces or tabs were found in the receptor "
                             "site file. However, DOCK 6 cannot properly "
                             "deal with them.")

        if "grid_score_grid_prefix" not in params:
            raise KeyError("The parameter 'grid_score_grid_prefix' has not "
                           "been defined in the input file.")
        elif (" " in params["grid_score_grid_prefix"]
                or "\t" in params["grid_score_grid_prefix"]):
            raise ValueError("Spaces or tabs were found in the grid file "
                             "prefix. However, DOCK 6 cannot properly "
                             "deal with them.")
        else:
            if not os.path.exists(params["grid_score_grid_prefix"] + ".bmp"):
                raise FileNotFoundError("The bmp grid file was not found in "
                                        "the working path.")
            if not os.path.exists(params["grid_score_grid_prefix"] + ".nrg"):
                raise FileNotFoundError("The nrg grid file was not found in "
                                        "the working path.")

    def dock(self, params_file, working_path):

        cur_path = os.path.abspath(os.getcwd())

        working_path = Path(working_path).absolute()
        os.chdir(working_path)

        output_file = f"{working_path}/dock6.out"

        params = Params.load(params_file)
        self._validate_params(params)

        try:
            command = f"{self.dock6} -i '{params_file}' -o '{output_file}'"

            p = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
            try:
                stdout, stderr = p.communicate()
            except Exception:
                p.kill()
                raise

            mol_file = params["ligand_outfile_prefix"] + "_scored.mol2"
            if not os.path.exists(mol_file):
                raise ProcessingFailed("The ligand output file was not "
                                       "created. Check the output file "
                                       "'%s' for more information."
                                       % output_file)
            elif os.path.getsize(mol_file) == 0:
                raise ProcessingFailed("The ligand output file is "
                                       "empty. Check the output file "
                                       "'%s' for more information."
                                       % output_file)

        except Exception as e:
            os.chdir(cur_path)
            logger.exception(e)
            raise



from itertools import product

from Bio.PDB.PDBIO import Select

from luna.interaction.cov import is_covalently_bound

import logging

logger = logging.getLogger()


class Neighbors:
    """Represents covalently bound neighbors of a residue."""

    def __init__(self, prev_res=None, next_res=None):

        self.predecessor = prev_res
        self.successor = next_res

    def has_predecessor(self):
        return self.predecessor is not None

    def has_successor(self):
        return self.successor is not None


def get_residue_neighbors(residue, select=Select(), verbose=True):
    """
    Identify covalently bound neighbors (previous/next) of a residue.

    In the case of an amino acid that is part of a peptide bond, its neighbors
    are any predecessor or successor residues. The same idea applies to nucleic acids.

    Parameters
    ----------
    residue : :class:`~luna.pdb.core.residue.Residue`
        The residue or other molecule from which covalently bound molecules
        will be recovered.
    select : :class:`~Bio.PDB.PDBIO.Select`
        Decides which atoms will be considered. By default,
        all atoms are accepted.
    verbose : bool
        If True, logs detailed debug info.

    Returns
    -------
    neighbors : dict of {str : :class:`~luna.pdb.core.residue.Residue`}
        A dictionary containing the predecessor (``previous``) or
        successor (``next``) molecules.
    """

    # Get valid atoms according to the provided selection function.
    trgt_res_atms = {atm.name: atm for atm in residue.get_atoms()
                     if select.accept_atom(atm)}

    if residue.is_residue():
        if "N" not in trgt_res_atms and verbose:
            logger.debug("There is a missing N in the residue %s. It may have "
                         "been filtered out by the provided selection "
                         "function. So, the predecessor residue cannot be "
                         "identified." % residue)
        if "C" not in trgt_res_atms and verbose:
            logger.debug("There is a missing C in the residue %s. It may have "
                         "been filtered out by the provided selection "
                         "function. So, the successor residue cannot be "
                         "identified." % residue)

        # If neither N and C are available in the valid atom list.
        if "N" not in trgt_res_atms and "C" not in trgt_res_atms:
            return {}

        neighbors = Neighbors()
        # If the chain has a residue coming before the target residue.
        if residue.idx - 1 >= 0:
            # First residue before the target in the chain list.
            prev_res = residue.parent.child_list[residue.idx - 1]
            # Get valid atoms according to the provided selection function.
            prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms()
                             if select.accept_atom(atm)}

            # A peptide bond exists between the C of one amino acid
            # and the N of another.
            if "C" in prev_res_atms and "N" in trgt_res_atms:
                if is_covalently_bound(trgt_res_atms["N"], prev_res_atms["C"]):
                    neighbors.predecessor = prev_res
                elif verbose:
                    logger.debug("The first residue before %s is too distant "
                                 "to fulfill the covalent bond threshold. It "
                                 "may be an indication of bad atom "
                                 "positioning or that there are missing "
                                 "residues." % residue)
            elif "C" not in prev_res_atms and verbose:
                logger.debug("There is a missing C in %s, the first residue "
                             "before %s in the chain list. It may have been "
                             "filtered out by the provided selection "
                             "function. So, the predecessor residue cannot "
                             "be identified." % (prev_res, residue))
        # Otherwise, it could mean that the residue is the first one in
        # the sequence or there are missing residues.
        elif verbose:
            logger.debug("The residue %s seems not to have any predecessor "
                         "residue. It may be the first in the chain sequence "
                         "or there are missing residues." % residue)

        # If the chain has a residue coming after the target residue.
        if residue.idx + 1 < len(residue.parent.child_list):
            # First residue after the target in the chain list.
            next_res = residue.parent.child_list[residue.idx + 1]
            # Get valid atoms according to the provided selection function.
            next_res_atms = {atm.name: atm for atm in next_res.get_atoms()
                             if select.accept_atom(atm)}

            # A peptide bond exists between the C of one amino acid and
            # the N of another.
            if "C" in trgt_res_atms and "N" in next_res_atms:
                if is_covalently_bound(trgt_res_atms["C"], next_res_atms["N"]):
                    neighbors.successor = next_res
                elif verbose:
                    logger.debug("The first residue after %s is too distant "
                                 "to fulfill the covalent thresholds. It may "
                                 "be an indication of bad atom positioning or "
                                 "that there are missing residues." % residue)
            elif "N" in next_res_atms and verbose:
                logger.debug("There is a missing N in %s, the first residue "
                             "after %s in the chain list. It may have been "
                             "filtered out by the provided selection "
                             "function. So, the predecessor residue cannot be "
                             "identified." % (next_res, residue))
        # Otherwise, it could mean that the residue is the last one in the
        # sequence or there are missing residues.
        elif verbose:
            logger.debug("The residue %s seems not to have any successor "
                         "residue. It may be the last in the chain sequence "
                         "or there are missing residues." % residue)
        return neighbors
    else:
        # First residue before the target in the chain list.
        prev_res = residue.parent.child_list[residue.idx - 1]
        # Get valid atoms according to the provided selection function.
        prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms()
                         if select.accept_atom(atm)}

        neighbors = Neighbors()
        # If the chain has a residue coming before the target residue.
        if residue.idx - 1 >= 0:
            # First residue before the target in the chain list.
            prev_res = residue.parent.child_list[residue.idx - 1]
            # Get valid atoms according to the provided selection function.
            prev_res_atms = {atm.name: atm for atm in prev_res.get_atoms()
                             if select.accept_atom(atm)}

            for trgt_atm, prev_atm in product(trgt_res_atms.values(),
                                              prev_res_atms.values()):
                if is_covalently_bound(trgt_atm, prev_atm):
                    neighbors.predecessor = prev_res
                    break

            if not neighbors.has_predecessor() and verbose:
                logger.debug("The first residue after %s is too distant to "
                             "fulfill the covalent thresholds. It may be an "
                             "indication of bad atom positioning or that "
                             "there are missing residues." % residue)
        # Otherwise, it could mean that the residue is the first one in the
        # sequence or there are missing residues.
        elif verbose:
            logger.debug("The residue %s seems not to have any predecessor "
                         "residue. It may be the first in the chain sequence "
                         "or there are missing residues." % residue)

        # If the chain has a residue coming after the target residue.
        if residue.idx + 1 < len(residue.parent.child_list):
            # First residue after the target in the chain list.
            next_res = residue.parent.child_list[residue.idx + 1]
            # Get valid atoms according to the provided selection function.
            next_res_atms = {atm.name: atm for atm in next_res.get_atoms()
                             if select.accept_atom(atm)}

            # Check each pair of atoms for covalently bonded atoms.
            for trgt_atm, next_atm in product(trgt_res_atms.values(),
                                              next_res_atms.values()):
                if is_covalently_bound(trgt_atm, next_atm):
                    neighbors.successor = next_res
                    break

            if not neighbors.has_successor() and verbose:
                logger.debug("The first residue after %s is too distant to "
                             "fulfill the covalent thresholds. It may be an "
                             "indication of bad atom positioning or that "
                             "there are missing residues." % residue)
        # Otherwise, it could mean that the residue is the last one in the
        # sequence or there are missing residues.
        elif verbose:
            logger.debug("The residue %s seems not to have any successor "
                         "residue. It may be the last in the chain sequence "
                         "or there are missing residues." % residue)
        return neighbors
    return {}
